@@ -21,18 +21,20 @@
 
 #include "collector.h"
 
-#include "value.h"
+#include "fast_malloc.h"
 #include "internal.h"
+#include "list.h"
+#include "value.h"
 
 #if APPLE_CHANGES
 #include <CoreFoundation/CoreFoundation.h>
 #include <cxxabi.h>
+#include <pthread.h>
+#include <mach/mach_port.h>
+#include <mach/task.h>
+#include <mach/thread_act.h>
 #endif
 
-#include <collector.h>
-#include <value.h>
-#include <internal.h>
-
 namespace KJS {
 
 // tunable parameters
@@ -56,7 +58,7 @@ struct CollectorCell {
     double memory[CELL_ARRAY_LENGTH];
     struct {
       void *zeroIfFree;
-      CollectorCell *next;
+      ptrdiff_t next;
     } freeCell;
   } u;
 };
@@ -79,7 +81,7 @@ struct CollectorHeap {
   int usedOversizeCells;
 
   int numLiveObjects;
-  int numAllocationsSinceLastCollect;
+  int numLiveObjectsAtLastCollect;
 };
 
 static CollectorHeap heap = {NULL, 0, 0, 0, NULL, 0, 0, 0, 0};
@@ -90,29 +92,29 @@ void* Collector::allocate(size_t s)
 {
   assert(Interpreter::lockCount() > 0);
 
-  if (s == 0)
-    return 0L;
-  
   // collect if needed
-  if (++heap.numAllocationsSinceLastCollect >= ALLOCATIONS_PER_COLLECTION) {
+  int numLiveObjects = heap.numLiveObjects;
+  if (numLiveObjects - heap.numLiveObjectsAtLastCollect >= ALLOCATIONS_PER_COLLECTION) {
     collect();
+    numLiveObjects = heap.numLiveObjects;
   }
   
-  if (s > (unsigned)CELL_SIZE) {
+  if (s > static_cast<size_t>(CELL_SIZE)) {
     // oversize allocator
     if (heap.usedOversizeCells == heap.numOversizeCells) {
       heap.numOversizeCells = MAX(MIN_ARRAY_SIZE, heap.numOversizeCells * GROWTH_FACTOR);
-      heap.oversizeCells = (CollectorCell **)realloc(heap.oversizeCells, heap.numOversizeCells * sizeof(CollectorCell *));
+      heap.oversizeCells = (CollectorCell **)kjs_fast_realloc(heap.oversizeCells, heap.numOversizeCells * sizeof(CollectorCell *));
     }
     
-    void *newCell = malloc(s);
+    void *newCell = kjs_fast_malloc(s);
     heap.oversizeCells[heap.usedOversizeCells] = (CollectorCell *)newCell;
     heap.usedOversizeCells++;
-    heap.numLiveObjects++;
+    heap.numLiveObjects = numLiveObjects + 1;
 
 #if !USE_CONSERVATIVE_GC
     ((ValueImp *)(newCell))->_flags = 0;
 #endif
+
     return newCell;
   }
   
@@ -135,10 +137,10 @@ void* Collector::allocate(size_t s)
     
     if (heap.usedBlocks == heap.numBlocks) {
       heap.numBlocks = MAX(MIN_ARRAY_SIZE, heap.numBlocks * GROWTH_FACTOR);
-      heap.blocks = (CollectorBlock **)realloc(heap.blocks, heap.numBlocks * sizeof(CollectorBlock *));
+      heap.blocks = (CollectorBlock **)kjs_fast_realloc(heap.blocks, heap.numBlocks * sizeof(CollectorBlock *));
     }
     
-    targetBlock = (CollectorBlock *)calloc(1, sizeof(CollectorBlock));
+    targetBlock = (CollectorBlock *)kjs_fast_calloc(1, sizeof(CollectorBlock));
     targetBlock->freeList = targetBlock->cells;
     heap.blocks[heap.usedBlocks] = targetBlock;
     heap.usedBlocks++;
@@ -146,33 +148,84 @@ void* Collector::allocate(size_t s)
   
   // find a free spot in the block and detach it from the free list
   CollectorCell *newCell = targetBlock->freeList;
-
-  if (newCell->u.freeCell.next != NULL) {
-    targetBlock->freeList = newCell->u.freeCell.next;
-  } else if (targetBlock->usedCells == (CELLS_PER_BLOCK - 1)) {
-    // last cell in this block
-    targetBlock->freeList = NULL;
-  } else {
-    // all next pointers are initially 0, meaning "next cell"
-    targetBlock->freeList = newCell + 1;
-  }
+  
+  // "next" field is a byte offset -- 0 means next cell, so a zeroed block is already initialized
+  // could avoid the casts by using a cell offset, but this avoids a relatively-slow multiply
+  targetBlock->freeList = reinterpret_cast<CollectorCell *>(reinterpret_cast<char *>(newCell + 1) + newCell->u.freeCell.next);
 
   targetBlock->usedCells++;
-  heap.numLiveObjects++;
+  heap.numLiveObjects = numLiveObjects + 1;
 
 #if !USE_CONSERVATIVE_GC
   ((ValueImp *)(newCell))->_flags = 0;
 #endif
-  return (void *)(newCell);
+
+  return newCell;
 }
 
 #if TEST_CONSERVATIVE_GC || USE_CONSERVATIVE_GC
 
+struct Collector::Thread {
+  Thread(pthread_t pthread, mach_port_t mthread) : posixThread(pthread), machThread(mthread) {}
+  Thread *next;
+  pthread_t posixThread;
+  mach_port_t machThread;
+};
+
+pthread_key_t registeredThreadKey;
+pthread_once_t registeredThreadKeyOnce = PTHREAD_ONCE_INIT;
+Collector::Thread *registeredThreads;
+  
+static void destroyRegisteredThread(void *data) 
+{
+  Collector::Thread *thread = (Collector::Thread *)data;
+
+  if (registeredThreads == thread) {
+    registeredThreads = registeredThreads->next;
+  } else {
+    Collector::Thread *last = registeredThreads;
+    for (Collector::Thread *t = registeredThreads->next; t != NULL; t = t->next) {
+      if (t == thread) {
+        last->next = t->next;
+          break;
+      }
+      last = t;
+    }
+  }
+
+  delete thread;
+}
+
+static void initializeRegisteredThreadKey()
+{
+  pthread_key_create(&registeredThreadKey, destroyRegisteredThread);
+}
+
+void Collector::registerThread()
+{
+  pthread_once(&registeredThreadKeyOnce, initializeRegisteredThreadKey);
+
+  if (!pthread_getspecific(registeredThreadKey)) {
+    pthread_t pthread = pthread_self();
+    Collector::Thread *thread = new Collector::Thread(pthread, pthread_mach_thread_np(pthread));
+    thread->next = registeredThreads;
+    registeredThreads = thread;
+    pthread_setspecific(registeredThreadKey, thread);
+  }
+}
+
+
 // cells are 8-byte aligned 
 #define IS_POINTER_ALIGNED(p) (((int)(p) & 7) == 0)
 
 void Collector::markStackObjectsConservatively(void *start, void *end)
 {
+  if (start > end) {
+    void *tmp = start;
+    start = end;
+    end = tmp;
+  }
+
   assert(((char *)end - (char *)start) < 0x1000000);
   assert(IS_POINTER_ALIGNED(start));
   assert(IS_POINTER_ALIGNED(end));
@@ -212,22 +265,74 @@ void Collector::markStackObjectsConservatively(void *start, void *end)
   }
 }
 
-void Collector::markStackObjectsConservatively()
+void Collector::markCurrentThreadConservatively()
 {
   jmp_buf registers;
   setjmp(registers);
 
   pthread_t thread = pthread_self();
   void *stackBase = pthread_get_stackaddr_np(thread);
-  void *stackPointer;
-  asm("mr %0,r1" : "=r" (stackPointer));
+  int dummy;
+  void *stackPointer = &dummy;
   markStackObjectsConservatively(stackPointer, stackBase);
 }
 
+typedef unsigned long usword_t; // word size, assumed to be either 32 or 64 bit
+
+void Collector::markOtherThreadConservatively(Thread *thread)
+{
+  thread_suspend(thread->machThread);
+
+#if defined(__i386__)
+  i386_thread_state_t regs;
+  unsigned user_count = sizeof(regs)/sizeof(int);
+  thread_state_flavor_t flavor = i386_THREAD_STATE;
+#elif defined(__ppc__)
+  ppc_thread_state_t  regs;
+  unsigned user_count = PPC_THREAD_STATE_COUNT;
+  thread_state_flavor_t flavor = PPC_THREAD_STATE;
+#elif defined(__ppc64__)
+  ppc_thread_state64_t  regs;
+  unsigned user_count = PPC_THREAD_STATE64_COUNT;
+  thread_state_flavor_t flavor = PPC_THREAD_STATE64;
+#else
+#error Unknown Architecture
+#endif
+  // get the thread register state
+  thread_get_state(thread->machThread, flavor, (thread_state_t)&regs, &user_count);
+  
+  // scan the registers
+  markStackObjectsConservatively((void *)&regs, (void *)((char *)&regs + (user_count * sizeof(usword_t))));
+  
+  // scan the stack
+#if defined(__i386__)
+  markStackObjectsConservatively((void *)regs.esp, pthread_get_stackaddr_np(thread->posixThread));
+#elif defined(__ppc__) || defined(__ppc64__)
+  markStackObjectsConservatively((void *)regs.r1, pthread_get_stackaddr_np(thread->posixThread));
+#else
+#error Unknown Architecture
+#endif
+
+  thread_resume(thread->machThread);
+}
+
+void Collector::markStackObjectsConservatively()
+{
+  markCurrentThreadConservatively();
+
+  for (Thread *thread = registeredThreads; thread != NULL; thread = thread->next) {
+    if (thread->posixThread != pthread_self()) {
+      markOtherThreadConservatively(thread);
+    }
+  }
+}
+
 void Collector::markProtectedObjects()
 {
-  for (int i = 0; i < ProtectedValues::_tableSize; i++) {
-    ValueImp *val = ProtectedValues::_table[i].key;
+  int size = ProtectedValues::_tableSize;
+  ProtectedValues::KeyValue *table = ProtectedValues::_table;
+  for (int i = 0; i < size; i++) {
+    ValueImp *val = table[i].key;
     if (val && !val->marked()) {
       val->mark();
     }
@@ -259,6 +364,7 @@ bool Collector::collect()
 
   markStackObjectsConservatively();
   markProtectedObjects();
+  List::markProtectedLists();
 #endif
 
 #if TEST_CONSERVATIVE_GC
@@ -315,20 +421,22 @@ bool Collector::collect()
   // SWEEP: delete everything with a zero refcount (garbage) and unmark everything else
   
   int emptyBlocks = 0;
+  int numLiveObjects = heap.numLiveObjects;
 
   for (int block = 0; block < heap.usedBlocks; block++) {
     CollectorBlock *curBlock = heap.blocks[block];
 
     int minimumCellsToProcess = curBlock->usedCells;
 
-    for (int cell = 0; cell < CELLS_PER_BLOCK; cell++) {
-      if (minimumCellsToProcess < cell) {
+    for (int i = 0; i < CELLS_PER_BLOCK; i++) {
+      if (minimumCellsToProcess < i) {
 	goto skip_block_sweep;
       }
 
-      ValueImp *imp = (ValueImp *)(curBlock->cells + cell);
+      CollectorCell *cell = curBlock->cells + i;
+      ValueImp *imp = reinterpret_cast<ValueImp *>(cell);
 
-      if (((CollectorCell *)imp)->u.freeCell.zeroIfFree != 0) {
+      if (cell->u.freeCell.zeroIfFree != 0) {
 #if USE_CONSERVATIVE_GC
 	if (!imp->_marked)
 #else
@@ -339,13 +447,13 @@ bool Collector::collect()
 	  // emulate destructing part of 'operator delete()'
 	  imp->~ValueImp();
 	  curBlock->usedCells--;
-	  heap.numLiveObjects--;
+	  numLiveObjects--;
 	  deleted = true;
 
 	  // put it on the free list
-	  ((CollectorCell *)imp)->u.freeCell.zeroIfFree = 0;
-	  ((CollectorCell *)imp)->u.freeCell.next = curBlock->freeList;
-	  curBlock->freeList = (CollectorCell *)imp;
+	  cell->u.freeCell.zeroIfFree = 0;
+	  cell->u.freeCell.next = reinterpret_cast<char *>(curBlock->freeList) - reinterpret_cast<char *>(cell + 1);
+	  curBlock->freeList = cell;
 
 	} else {
 #if USE_CONSERVATIVE_GC
@@ -367,7 +475,7 @@ bool Collector::collect()
       emptyBlocks++;
       if (emptyBlocks > SPARE_EMPTY_BLOCKS) {
 #if !DEBUG_COLLECTOR
-	free(heap.blocks[block]);
+	kjs_fast_free(heap.blocks[block]);
 #endif
 	// swap with the last block so we compact as we go
 	heap.blocks[block] = heap.blocks[heap.usedBlocks - 1];
@@ -376,7 +484,7 @@ bool Collector::collect()
 
 	if (heap.numBlocks > MIN_ARRAY_SIZE && heap.usedBlocks < heap.numBlocks / LOW_WATER_FACTOR) {
 	  heap.numBlocks = heap.numBlocks / GROWTH_FACTOR; 
-	  heap.blocks = (CollectorBlock **)realloc(heap.blocks, heap.numBlocks * sizeof(CollectorBlock *));
+	  heap.blocks = (CollectorBlock **)kjs_fast_realloc(heap.blocks, heap.numBlocks * sizeof(CollectorBlock *));
 	}
       } 
     }
@@ -401,7 +509,7 @@ bool Collector::collect()
 #if DEBUG_COLLECTOR
       heap.oversizeCells[cell]->u.freeCell.zeroIfFree = 0;
 #else
-      free((void *)imp);
+      kjs_fast_free((void *)imp);
 #endif
 
       // swap with the last oversize cell so we compact as we go
@@ -409,11 +517,11 @@ bool Collector::collect()
 
       heap.usedOversizeCells--;
       deleted = true;
-      heap.numLiveObjects--;
+      numLiveObjects--;
 
       if (heap.numOversizeCells > MIN_ARRAY_SIZE && heap.usedOversizeCells < heap.numOversizeCells / LOW_WATER_FACTOR) {
 	heap.numOversizeCells = heap.numOversizeCells / GROWTH_FACTOR; 
-	heap.oversizeCells = (CollectorCell **)realloc(heap.oversizeCells, heap.numOversizeCells * sizeof(CollectorCell *));
+	heap.oversizeCells = (CollectorCell **)kjs_fast_realloc(heap.oversizeCells, heap.numOversizeCells * sizeof(CollectorCell *));
       }
 
     } else {
@@ -428,9 +536,10 @@ bool Collector::collect()
     }
   }
   
-  heap.numAllocationsSinceLastCollect = 0;
+  heap.numLiveObjects = numLiveObjects;
+  heap.numLiveObjectsAtLastCollect = numLiveObjects;
   
-  memoryFull = (heap.numLiveObjects >= KJS_MEM_LIMIT);
+  memoryFull = (numLiveObjects >= KJS_MEM_LIMIT);
 
   return deleted;
 }
@@ -494,8 +603,10 @@ int Collector::numReferencedObjects()
   int count = 0;
 
 #if USE_CONSERVATIVE_GC
-  for (int i = 0; i < ProtectedValues::_tableSize; i++) {
-    ValueImp *val = ProtectedValues::_table[i].key;
+  int size = ProtectedValues::_tableSize;
+  ProtectedValues::KeyValue *table = ProtectedValues::_table;
+  for (int i = 0; i < size; i++) {
+    ValueImp *val = table[i].key;
     if (val) {
       ++count;
     }
@@ -532,8 +643,10 @@ const void *Collector::rootObjectClasses()
   CFMutableSetRef classes = CFSetCreateMutable(NULL, 0, &kCFTypeSetCallBacks);
 
 #if USE_CONSERVATIVE_GC
-  for (int i = 0; i < ProtectedValues::_tableSize; i++) {
-    ValueImp *val = ProtectedValues::_table[i].key;
+  int size = ProtectedValues::_tableSize;
+  ProtectedValues::KeyValue *table = ProtectedValues::_table;
+  for (int i = 0; i < size; i++) {
+    ValueImp *val = table[i].key;
     if (val) {
       const char *mangled_name = typeid(*val).name();
       int status;
