@@ -47,15 +47,25 @@
 
 using namespace KJS;
 
+// Disabled for now because it shows up on benchmark (0.5%).
+#if DEBUGGER_SUPPORT
+
 #define KJS_BREAKPOINT \
   if (!hitStatement(exec)) \
     return Completion(Normal);
 
 #define KJS_ABORTPOINT \
-  if (exec->interpreter()->imp()->debugger() && \
-      exec->interpreter()->imp()->debugger()->imp()->aborted()) \
+  if (exec->dynamicInterpreter()->imp()->debugger() && \
+      exec->dynamicInterpreter()->imp()->debugger()->imp()->aborted()) \
     return Completion(Normal);
 
+#else
+
+#define KJS_BREAKPOINT
+#define KJS_ABORTPOINT
+
+#endif
+
 #define KJS_CHECKEXCEPTION \
   if (exec->hadException()) \
     return Completion(Throw, exec->exception()); \
@@ -89,6 +99,7 @@ std::list<Node *> * Node::s_nodes = 0L;
 Node::Node()
 {
   line = Lexer::curr()->lineNo();
+  sourceURL = Lexer::curr()->sourceURL();
   refcount = 0;
 #ifdef KJS_DEBUG_MEM
   if (!s_nodes)
@@ -125,7 +136,7 @@ void Node::finalCheck()
 
 Value Node::throwError(ExecState *exec, ErrorType e, const char *msg)
 {
-  Object err = Error::create(exec, e, msg, lineNo(), sourceId());
+  Object err = Error::create(exec, e, msg, lineNo(), sourceId(), &sourceURL);
   exec->setException(err);
   return err;
 }
@@ -176,7 +187,7 @@ void StatementNode::setLoc(int line0, int line1, int sourceId)
 // return true if the debugger wants us to stop at this point
 bool StatementNode::hitStatement(ExecState *exec)
 {
-  Debugger *dbg = exec->interpreter()->imp()->debugger();
+  Debugger *dbg = exec->dynamicInterpreter()->imp()->debugger();
   if (dbg)
     return dbg->atStatement(exec,sid,l0,l1);
   else
@@ -186,7 +197,7 @@ bool StatementNode::hitStatement(ExecState *exec)
 // return true if the debugger wants us to stop at this point
 bool StatementNode::abortStatement(ExecState *exec)
 {
-  Debugger *dbg = exec->interpreter()->imp()->debugger();
+  Debugger *dbg = exec->dynamicInterpreter()->imp()->debugger();
   if (dbg)
     return dbg->imp()->aborted();
   else
@@ -208,21 +219,21 @@ Value NullNode::evaluate(ExecState */*exec*/)
 
 Value BooleanNode::evaluate(ExecState */*exec*/)
 {
-  return Boolean(value);
+  return Value(value);
 }
 
 // ------------------------------ NumberNode -----------------------------------
 
 Value NumberNode::evaluate(ExecState */*exec*/)
 {
-  return Number(value);
+  return Value(value);
 }
 
 // ------------------------------ StringNode -----------------------------------
 
 Value StringNode::evaluate(ExecState */*exec*/)
 {
-  return String(value);
+  return value;
 }
 
 // ------------------------------ RegExpNode -----------------------------------
@@ -235,7 +246,7 @@ Value RegExpNode::evaluate(ExecState *exec)
   list.append(p);
   list.append(f);
 
-  Object reg = exec->interpreter()->imp()->builtinRegExp();
+  Object reg = exec->lexicalInterpreter()->imp()->builtinRegExp();
   return reg.construct(exec,list);
 }
 
@@ -301,6 +312,11 @@ Value GroupNode::evaluate(ExecState *exec)
   return group->evaluate(exec);
 }
 
+Reference GroupNode::evaluateReference(ExecState *exec)
+{
+  return group->evaluateReference(exec);
+}
+
 // ------------------------------ ElementNode ----------------------------------
 
 void ElementNode::ref()
@@ -328,7 +344,7 @@ bool ElementNode::deref()
 // ECMA 11.1.4
 Value ElementNode::evaluate(ExecState *exec)
 {
-  Object array = exec->interpreter()->builtinArray().construct(exec, List::empty());
+  Object array = exec->lexicalInterpreter()->builtinArray().construct(exec, List::empty());
   int length = 0;
   for (ElementNode *n = this; n; n = n->list) {
     Value val = n->node->evaluate(exec);
@@ -366,13 +382,13 @@ Value ArrayNode::evaluate(ExecState *exec)
     KJS_CHECKEXCEPTIONVALUE
     length = opt ? array.get(exec,lengthPropertyName).toInt32(exec) : 0;
   } else {
-    Value newArr = exec->interpreter()->builtinArray().construct(exec,List::empty());
+    Value newArr = exec->lexicalInterpreter()->builtinArray().construct(exec,List::empty());
     array = Object(static_cast<ObjectImp*>(newArr.imp()));
     length = 0;
   }
 
   if (opt)
-    array.put(exec,lengthPropertyName, Number(elision + length), DontEnum | DontDelete);
+    array.put(exec,lengthPropertyName, Value(elision + length), DontEnum | DontDelete);
 
   return array;
 }
@@ -399,7 +415,7 @@ Value ObjectLiteralNode::evaluate(ExecState *exec)
   if (list)
     return list->evaluate(exec);
 
-  return exec->interpreter()->builtinObject().construct(exec,List::empty());
+  return exec->lexicalInterpreter()->builtinObject().construct(exec,List::empty());
 }
 
 // ------------------------------ PropertyValueNode ----------------------------
@@ -433,7 +449,7 @@ bool PropertyValueNode::deref()
 // ECMA 11.1.5
 Value PropertyValueNode::evaluate(ExecState *exec)
 {
-  Object obj = exec->interpreter()->builtinObject().construct(exec, List::empty());
+  Object obj = exec->lexicalInterpreter()->builtinObject().construct(exec, List::empty());
   
   for (PropertyValueNode *p = this; p; p = p->list) {
     Value n = p->name->evaluate(exec);
@@ -689,40 +705,32 @@ Value FunctionCallNode::evaluate(ExecState *exec)
   Value v = ref.getValue(exec);
 
   if (v.type() != ObjectType) {
-    return throwError(exec, TypeError, "Value %s (result of expression %s) is not object. Cannot be called.", v, expr);
+    return throwError(exec, TypeError, "Value %s (result of expression %s) is not object.", v, expr);
   }
 
-  Object func = Object(static_cast<ObjectImp*>(v.imp()));
+  ObjectImp *func = static_cast<ObjectImp*>(v.imp());
 
-  if (!func.implementsCall()) {
+  if (!func->implementsCall()) {
     return throwError(exec, TypeError, "Object %s (result of expression %s) does not allow calls.", v, expr);
   }
 
-  Value thisVal;
-  if (ref.isMutable())
-    thisVal = ref.getBase(exec);
-  else
-    thisVal = Null();
+  ObjectImp *thisObjImp = 0;
+  ValueImp *thisValImp = ref.baseIfMutable();
+  if (thisValImp && thisValImp->type() == ObjectType && !static_cast<ObjectImp *>(thisValImp)->inherits(&ActivationImp::info))
+    thisObjImp = static_cast<ObjectImp *>(thisValImp);
 
-  if (thisVal.type() == ObjectType &&
-      Object::dynamicCast(thisVal).inherits(&ActivationImp::info))
-    thisVal = Null();
-
-  if (thisVal.type() != ObjectType) {
+  if (!thisObjImp) {
     // ECMA 11.2.3 says that in this situation the this value should be null.
     // However, section 10.2.3 says that in the case where the value provided
     // by the caller is null, the global object should be used. It also says
     // that the section does not apply to interal functions, but for simplicity
     // of implementation we use the global object anyway here. This guarantees
     // that in host objects you always get a valid object for this.
-    // thisVal = Null();
-    thisVal = exec->interpreter()->globalObject();
+    thisObjImp = exec->dynamicInterpreter()->globalObject().imp();
   }
 
-  Object thisObj = Object::dynamicCast(thisVal);
-  Value result = func.call(exec,thisObj, argList);
-
-  return result;
+  Object thisObj(thisObjImp);
+  return func->call(exec, thisObj, argList);
 }
 
 // ------------------------------ PostfixNode ----------------------------------
@@ -747,14 +755,14 @@ Value PostfixNode::evaluate(ExecState *exec)
   Reference ref = expr->evaluateReference(exec);
   KJS_CHECKEXCEPTIONVALUE
   Value v = ref.getValue(exec);
-  Number n = v.toNumber(exec);
 
-  double newValue = (oper == OpPlusPlus) ? n.value() + 1 : n.value() - 1;
-  Value n2 = Number(newValue);
+  bool knownToBeInteger;
+  double n = v.toNumber(exec, knownToBeInteger);
 
-  ref.putValue(exec,n2);
+  double newValue = (oper == OpPlusPlus) ? n + 1 : n - 1;
+  ref.putValue(exec, Value(newValue, knownToBeInteger));
 
-  return n;
+  return Value(n, knownToBeInteger);
 }
 
 // ------------------------------ DeleteNode -----------------------------------
@@ -778,7 +786,7 @@ Value DeleteNode::evaluate(ExecState *exec)
 {
   Reference ref = expr->evaluateReference(exec);
   KJS_CHECKEXCEPTIONVALUE
-  return Boolean(ref.deleteValue(exec));
+  return Value(ref.deleteValue(exec));
 }
 
 // ------------------------------ VoidNode -------------------------------------
@@ -828,11 +836,9 @@ Value TypeOfNode::evaluate(ExecState *exec)
   const char *s = 0L;
   Reference ref = expr->evaluateReference(exec);
   KJS_CHECKEXCEPTIONVALUE
-  if (ref.isMutable()) {
-    Value b = ref.getBase(exec);
-    if (b.type() == NullType)
-      return String("undefined");
-  }
+  ValueImp *b = ref.baseIfMutable();
+  if (b && b->dispatchType() == NullType)
+    return Value("undefined");
   Value v = ref.getValue(exec);
   switch (v.type())
     {
@@ -859,7 +865,7 @@ Value TypeOfNode::evaluate(ExecState *exec)
       break;
     }
 
-  return String(s);
+  return Value(s);
 }
 
 // ------------------------------ PrefixNode -----------------------------------
@@ -884,12 +890,14 @@ Value PrefixNode::evaluate(ExecState *exec)
   Reference ref = expr->evaluateReference(exec);
   KJS_CHECKEXCEPTIONVALUE
   Value v = ref.getValue(exec);
-  Number n = v.toNumber(exec);
 
-  double newValue = (oper == OpPlusPlus) ? n.value() + 1 : n.value() - 1;
-  Value n2 = Number(newValue);
+  bool knownToBeInteger;
+  double n = v.toNumber(exec, knownToBeInteger);
 
-  ref.putValue(exec,n2);
+  double newValue = (oper == OpPlusPlus) ? n + 1 : n - 1;
+  Value n2(newValue, knownToBeInteger);
+
+  ref.putValue(exec, n2);
 
   return n2;
 }
@@ -916,7 +924,7 @@ Value UnaryPlusNode::evaluate(ExecState *exec)
   Value v = expr->evaluate(exec);
   KJS_CHECKEXCEPTIONVALUE
 
-  return Number(v.toNumber(exec)); /* TODO: optimize */
+  return Value(v.toNumber(exec)); /* TODO: optimize */
 }
 
 // ------------------------------ NegateNode -----------------------------------
@@ -940,11 +948,10 @@ Value NegateNode::evaluate(ExecState *exec)
 {
   Value v = expr->evaluate(exec);
   KJS_CHECKEXCEPTIONVALUE
-  Number n = v.toNumber(exec);
 
-  double d = -n.value();
-
-  return Number(d);
+  bool knownToBeInteger;
+  double n = v.toNumber(exec, knownToBeInteger);
+  return Value(-n, knownToBeInteger && n != 0);
 }
 
 // ------------------------------ BitwiseNotNode -------------------------------
@@ -968,9 +975,7 @@ Value BitwiseNotNode::evaluate(ExecState *exec)
 {
   Value v = expr->evaluate(exec);
   KJS_CHECKEXCEPTIONVALUE
-  int i32 = v.toInt32(exec);
-
-  return Number(~i32);
+  return Value(~v.toInt32(exec));
 }
 
 // ------------------------------ LogicalNotNode -------------------------------
@@ -994,9 +999,7 @@ Value LogicalNotNode::evaluate(ExecState *exec)
 {
   Value v = expr->evaluate(exec);
   KJS_CHECKEXCEPTIONVALUE
-  bool b = v.toBoolean(exec);
-
-  return Boolean(!b);
+  return Value(!v.toBoolean(exec));
 }
 
 // ------------------------------ MultNode -------------------------------------
@@ -1028,7 +1031,7 @@ Value MultNode::evaluate(ExecState *exec)
   Value v2 = term2->evaluate(exec);
   KJS_CHECKEXCEPTIONVALUE
 
-  return mult(exec,v1, v2, oper);
+  return mult(exec, v1, v2, oper);
 }
 
 // ------------------------------ AddNode --------------------------------------
@@ -1060,7 +1063,7 @@ Value AddNode::evaluate(ExecState *exec)
   Value v2 = term2->evaluate(exec);
   KJS_CHECKEXCEPTIONVALUE
 
-  return add(exec,v1, v2, oper);
+  return add(exec, v1, v2, oper);
 }
 
 // ------------------------------ ShiftNode ------------------------------------
@@ -1093,23 +1096,17 @@ Value ShiftNode::evaluate(ExecState *exec)
   unsigned int i2 = v2.toUInt32(exec);
   i2 &= 0x1f;
 
-  long result;
   switch (oper) {
   case OpLShift:
-    result = v1.toInt32(exec) << i2;
-    break;
+    return Value(v1.toInt32(exec) << i2);
   case OpRShift:
-    result = v1.toInt32(exec) >> i2;
-    break;
+    return Value(v1.toInt32(exec) >> i2);
   case OpURShift:
-    result = v1.toUInt32(exec) >> i2;
-    break;
+    return Value(v1.toUInt32(exec) >> i2);
   default:
     assert(!"ShiftNode: unhandled switch case");
-    result = 0L;
+    return Undefined();
   }
-
-  return Number(static_cast<double>(result));
 }
 
 // ------------------------------ RelationalNode -------------------------------
@@ -1171,14 +1168,14 @@ Value RelationalNode::evaluate(ExecState *exec)
       // But we are supposed to throw an exception where the object does not "have" the [[HasInstance]]
       // property. It seems that all object have the property, but not all implement it, so in this
       // case we return false (consistent with mozilla)
-      return Boolean(false);
+      return Value(false);
       //      return throwError(exec, TypeError,
       //			"Object does not implement the [[HasInstance]] method." );
     }
     return o2.hasInstance(exec, v1);
   }
 
-  return Boolean(b);
+  return Value(b);
 }
 
 // ------------------------------ EqualNode ------------------------------------
@@ -1219,7 +1216,7 @@ Value EqualNode::evaluate(ExecState *exec)
     bool eq = strictEqual(exec,v1, v2);
     result = oper == OpStrEq ? eq : !eq;
   }
-  return Boolean(result);
+  return Value(result);
 }
 
 // ------------------------------ BitOperNode ----------------------------------
@@ -1259,7 +1256,7 @@ Value BitOperNode::evaluate(ExecState *exec)
   else
     result = i1 | i2;
 
-  return Number(result);
+  return Value(result);
 }
 
 // ------------------------------ BinaryLogicalNode ----------------------------
@@ -1370,8 +1367,8 @@ Value AssignNode::evaluate(ExecState *exec)
     Value v1 = l.getValue(exec);
     Value v2 = expr->evaluate(exec);
     KJS_CHECKEXCEPTIONVALUE
-    int i1 = v1.toInt32(exec);
-    int i2 = v2.toInt32(exec);
+    int i1;
+    int i2;
     unsigned int ui;
     switch (oper) {
     case OpMultEq:
@@ -1387,28 +1384,41 @@ Value AssignNode::evaluate(ExecState *exec)
       v = add(exec, v1, v2, '-');
       break;
     case OpLShift:
-      v = Number(i1 <<= i2);
+      i1 = v1.toInt32(exec);
+      i2 = v2.toInt32(exec);
+      v = Value(i1 << i2);
       break;
     case OpRShift:
-      v = Number(i1 >>= i2);
+      i1 = v1.toInt32(exec);
+      i2 = v2.toInt32(exec);
+      v = Value(i1 >> i2);
       break;
     case OpURShift:
       ui = v1.toUInt32(exec);
-      v = Number(ui >>= i2);
+      i2 = v2.toInt32(exec);
+      v = Value(ui >> i2);
       break;
     case OpAndEq:
-      v = Number(i1 &= i2);
+      i1 = v1.toInt32(exec);
+      i2 = v2.toInt32(exec);
+      v = Value(i1 & i2);
       break;
     case OpXOrEq:
-      v = Number(i1 ^= i2);
+      i1 = v1.toInt32(exec);
+      i2 = v2.toInt32(exec);
+      v = Value(i1 ^ i2);
       break;
     case OpOrEq:
-      v = Number(i1 |= i2);
+      i1 = v1.toInt32(exec);
+      i2 = v2.toInt32(exec);
+      v = Value(i1 | i2);
       break;
     case OpModEq: {
-      double d1 = v1.toNumber(exec);
-      double d2 = v2.toNumber(exec);
-      v = Number(fmod(d1,d2));
+      bool d1KnownToBeInteger;
+      double d1 = v1.toNumber(exec, d1KnownToBeInteger);
+      bool d2KnownToBeInteger;
+      double d2 = v2.toNumber(exec, d2KnownToBeInteger);
+      v = Value(fmod(d1, d2), d1KnownToBeInteger && d2KnownToBeInteger && d2 != 0);
     }
       break;
     default:
@@ -1578,7 +1588,7 @@ bool VarDeclNode::deref()
 // ECMA 12.2
 Value VarDeclNode::evaluate(ExecState *exec)
 {
-  Object variable = Object::dynamicCast(exec->context().imp()->variableObject());
+  Object variable = exec->context().imp()->variableObject();
 
   Value val;
   if (init) {
@@ -1599,7 +1609,7 @@ Value VarDeclNode::evaluate(ExecState *exec)
   // "var location" creates a dynamic property instead of activating window.location.
   variable.put(exec, ident, val, DontDelete | Internal);
 
-  return String(ident.ustring());
+  return ident.ustring();
 }
 
 void VarDeclNode::processVarDecls(ExecState *exec)
@@ -1960,8 +1970,7 @@ bool ForNode::deref()
 // ECMA 12.6.3
 Completion ForNode::execute(ExecState *exec)
 {
-  Value e, v, cval;
-  bool b;
+  Value v, cval;
 
   if (expr1) {
     v = expr1->evaluate(exec);
@@ -1971,8 +1980,7 @@ Completion ForNode::execute(ExecState *exec)
     if (expr2) {
       v = expr2->evaluate(exec);
       KJS_CHECKEXCEPTION
-      b = v.toBoolean(exec);
-      if (b == false)
+      if (!v.toBoolean(exec))
 	return Completion(Normal, cval);
     }
     // bail out on error
@@ -2725,9 +2733,9 @@ void FuncDeclNode::processFuncDecl(ExecState *exec)
   FunctionImp *fimp = new DeclaredFunctionImp(exec, ident, body, exec->context().imp()->scopeChain());
   Object func(fimp); // protect from GC
 
-  //  Value proto = exec->interpreter()->builtinObject().construct(exec,List::empty());
+  //  Value proto = exec->lexicalInterpreter()->builtinObject().construct(exec,List::empty());
   List empty;
-  Object proto = exec->interpreter()->builtinObject().construct(exec,empty);
+  Object proto = exec->lexicalInterpreter()->builtinObject().construct(exec,empty);
   proto.put(exec, constructorPropertyName, func, ReadOnly|DontDelete|DontEnum);
   func.put(exec, prototypePropertyName, proto, Internal|DontDelete);
 
@@ -2778,7 +2786,7 @@ Value FuncExprNode::evaluate(ExecState *exec)
   FunctionImp *fimp = new DeclaredFunctionImp(exec, Identifier::null(), body, exec->context().imp()->scopeChain());
   Value ret(fimp);
   List empty;
-  Value proto = exec->interpreter()->builtinObject().construct(exec,empty);
+  Value proto = exec->lexicalInterpreter()->builtinObject().construct(exec,empty);
   fimp->put(exec, prototypePropertyName, proto, Internal|DontDelete);
 
   int plen = 0;
