@@ -25,6 +25,8 @@
 #ifndef _NODES_H_
 #define _NODES_H_
 
+#include "fast_malloc.h"
+
 #include "internal.h"
 //#include "debugger.h"
 #ifndef NDEBUG
@@ -77,6 +79,9 @@ namespace KJS {
   public:
     Node();
     virtual ~Node();
+
+    KJS_FAST_ALLOCATED;
+
     virtual Value evaluate(ExecState *exec) = 0;
     virtual Reference evaluateReference(ExecState *exec);
     UString toString() const;
@@ -202,7 +207,8 @@ namespace KJS {
     GroupNode(Node *g) : group(g) { }
     virtual void ref();
     virtual bool deref();
-    Value evaluate(ExecState *exec);
+    virtual Value evaluate(ExecState *exec);
+    virtual Reference evaluateReference(ExecState *exec);
     virtual void streamTo(SourceStream &s) const { group->streamTo(s); }
   private:
     Node *group;
