@@ -1,6 +1,6 @@
 /*
  *  This file is part of the KDE libraries
- *  Copyright (C) 2003 Apple Computer, Inc.
+ *  Copyright (C) 2004 Apple Computer, Inc.
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Library General Public
@@ -21,6 +21,7 @@
 
 #include "property_map.h"
 
+#include "fast_malloc.h"
 #include "object.h"
 #include "protect.h"
 #include "reference_list.h"
@@ -76,6 +77,7 @@ struct PropertyMapHashTable
     int sizeMask;
     int size;
     int keyCount;
+    int sentinelCount;
     int lastIndexUsed;
     PropertyMapHashTableEntry entries[1];
 };
@@ -111,12 +113,14 @@ PropertyMap::~PropertyMap()
         return;
     }
     
-    for (int i = 0; i < _table->size; i++) {
-        UString::Rep *key = _table->entries[i].key;
+    int size = _table->size;
+    Entry *entries = _table->entries;
+    for (int i = 0; i < size; i++) {
+        UString::Rep *key = entries[i].key;
         if (key)
             key->deref();
     }
-    free(_table);
+    kjs_fast_free(_table);
 }
 
 void PropertyMap::clear()
@@ -132,14 +136,18 @@ void PropertyMap::clear()
         return;
     }
 
-    for (int i = 0; i < _table->size; i++) {
-        UString::Rep *key = _table->entries[i].key;
+    int size = _table->size;
+    Entry *entries = _table->entries;
+    for (int i = 0; i < size; i++) {
+        UString::Rep *key = entries[i].key;
         if (key) {
             key->deref();
-            _table->entries[i].key = 0;
+            entries[i].key = 0;
+            entries[i].value = 0;
         }
     }
     _table->keyCount = 0;
+    _table->sentinelCount = 0;
 }
 
 ValueImp *PropertyMap::get(const Identifier &name, int &attributes) const
@@ -160,20 +168,22 @@ ValueImp *PropertyMap::get(const Identifier &name, int &attributes) const
     }
     
     unsigned h = rep->hash();
-    int i = h & _table->sizeMask;
+    int sizeMask = _table->sizeMask;
+    Entry *entries = _table->entries;
+    int i = h & sizeMask;
     int k = 0;
 #if DUMP_STATISTICS
     ++numProbes;
-    numCollisions += _table->entries[i].key && _table->entries[i].key != rep;
+    numCollisions += entries[i].key && entries[i].key != rep;
 #endif
-    while (UString::Rep *key = _table->entries[i].key) {
+    while (UString::Rep *key = entries[i].key) {
         if (rep == key) {
-            attributes = _table->entries[i].attributes;
-            return _table->entries[i].value;
+            attributes = entries[i].attributes;
+            return entries[i].value;
         }
         if (k == 0)
-            k = 1 | (h % _table->sizeMask);
-        i = (i + k) & _table->sizeMask;
+            k = 1 | (h % sizeMask);
+        i = (i + k) & sizeMask;
 #if DUMP_STATISTICS
         ++numRehashes;
 #endif
@@ -197,18 +207,20 @@ ValueImp *PropertyMap::get(const Identifier &name) const
     }
     
     unsigned h = rep->hash();
-    int i = h & _table->sizeMask;
+    int sizeMask = _table->sizeMask;
+    Entry *entries = _table->entries;
+    int i = h & sizeMask;
     int k = 0;
 #if DUMP_STATISTICS
     ++numProbes;
-    numCollisions += _table->entries[i].key && _table->entries[i].key != rep;
+    numCollisions += entries[i].key && entries[i].key != rep;
 #endif
-    while (UString::Rep *key = _table->entries[i].key) {
+    while (UString::Rep *key = entries[i].key) {
         if (rep == key)
-            return _table->entries[i].value;
+            return entries[i].value;
         if (k == 0)
-            k = 1 | (h % _table->sizeMask);
-        i = (i + k) & _table->sizeMask;
+            k = 1 | (h % sizeMask);
+        i = (i + k) & sizeMask;
 #if DUMP_STATISTICS
         ++numRehashes;
 #endif
@@ -274,38 +286,49 @@ void PropertyMap::put(const Identifier &name, ValueImp *value, int attributes)
         expand();
     
     unsigned h = rep->hash();
-    int i = h & _table->sizeMask;
+    int sizeMask = _table->sizeMask;
+    Entry *entries = _table->entries;
+    int i = h & sizeMask;
     int k = 0;
+    bool foundDeletedElement = false;
+    int deletedElementIndex = 0;    /* initialize to make the compiler happy */
 #if DUMP_STATISTICS
     ++numProbes;
-    numCollisions += _table->entries[i].key && _table->entries[i].key != rep;
+    numCollisions += entries[i].key && entries[i].key != rep;
 #endif
-    while (UString::Rep *key = _table->entries[i].key) {
+    while (UString::Rep *key = entries[i].key) {
         if (rep == key) {
             // Put a new value in an existing hash table entry.
-            _table->entries[i].value = value;
+            entries[i].value = value;
             // Attributes are intentionally not updated.
             return;
         }
-        // If we find the deleted-element sentinel, insert on top of it.
-        if (key == &UString::Rep::null) {
-            key->deref();
-            break;
+        // If we find the deleted-element sentinel, remember it for use later.
+        if (key == &UString::Rep::null && !foundDeletedElement) {
+            foundDeletedElement = true;
+            deletedElementIndex = i;
         }
         if (k == 0)
-            k = 1 | (h % _table->sizeMask);
-        i = (i + k) & _table->sizeMask;
+            k = 1 | (h % sizeMask);
+        i = (i + k) & sizeMask;
 #if DUMP_STATISTICS
         ++numRehashes;
 #endif
     }
-    
+
+    // Use either the deleted element or the 0 at the end of the chain.
+    if (foundDeletedElement) {
+        i = deletedElementIndex;
+        entries[i].key->deref();
+        --_table->sentinelCount;
+    }
+
     // Create a new hash table entry.
     rep->ref();
-    _table->entries[i].key = rep;
-    _table->entries[i].value = value;
-    _table->entries[i].attributes = attributes;
-    _table->entries[i].index = ++_table->lastIndexUsed;
+    entries[i].key = rep;
+    entries[i].value = value;
+    entries[i].attributes = attributes;
+    entries[i].index = ++_table->lastIndexUsed;
     ++_table->keyCount;
 
     checkConsistency();
@@ -316,29 +339,45 @@ void PropertyMap::insert(UString::Rep *key, ValueImp *value, int attributes, int
     assert(_table);
 
     unsigned h = key->hash();
-    int i = h & _table->sizeMask;
+    int sizeMask = _table->sizeMask;
+    Entry *entries = _table->entries;
+    int i = h & sizeMask;
     int k = 0;
 #if DUMP_STATISTICS
     ++numProbes;
-    numCollisions += _table->entries[i].key && _table->entries[i].key != key;
+    numCollisions += entries[i].key && entries[i].key != key;
 #endif
-    while (_table->entries[i].key) {
-        assert(_table->entries[i].key != &UString::Rep::null);
+    while (entries[i].key) {
+        assert(entries[i].key != &UString::Rep::null);
         if (k == 0)
-            k = 1 | (h % _table->sizeMask);
-        i = (i + k) & _table->sizeMask;
+            k = 1 | (h % sizeMask);
+        i = (i + k) & sizeMask;
 #if DUMP_STATISTICS
         ++numRehashes;
 #endif
     }
     
-    _table->entries[i].key = key;
-    _table->entries[i].value = value;
-    _table->entries[i].attributes = attributes;
-    _table->entries[i].index = index;
+    entries[i].key = key;
+    entries[i].value = value;
+    entries[i].attributes = attributes;
+    entries[i].index = index;
 }
 
 void PropertyMap::expand()
+{
+    Table *oldTable = _table;
+    int oldTableSize = oldTable ? oldTable->size : 0;    
+    rehash(oldTableSize ? oldTableSize * 2 : 16);
+}
+
+void PropertyMap::rehash()
+{
+    assert(_table);
+    assert(_table->size);
+    rehash(_table->size);
+}
+
+void PropertyMap::rehash(int newTableSize)
 {
     checkConsistency();
     
@@ -346,8 +385,7 @@ void PropertyMap::expand()
     int oldTableSize = oldTable ? oldTable->size : 0;
     int oldTableKeyCount = oldTable ? oldTable->keyCount : 0;
     
-    int newTableSize = oldTableSize ? oldTableSize * 2 : 16;
-    _table = (Table *)calloc(1, sizeof(Table) + (newTableSize - 1) * sizeof(Entry) );
+    _table = (Table *)kjs_fast_calloc(1, sizeof(Table) + (newTableSize - 1) * sizeof(Entry) );
     _table->size = newTableSize;
     _table->sizeMask = newTableSize - 1;
     _table->keyCount = oldTableKeyCount;
@@ -357,8 +395,8 @@ void PropertyMap::expand()
     if (key) {
         insert(key, _singleEntry.value, _singleEntry.attributes, 0);
         _singleEntry.key = 0;
-	// update the count, because single entries don't count towards
-	// the table key count
+        // update the count, because single entries don't count towards
+        // the table key count
 	++_table->keyCount;
 	assert(_table->keyCount == 1);
     }
@@ -381,7 +419,7 @@ void PropertyMap::expand()
     }
     _table->lastIndexUsed = lastIndexUsed;
 
-    free(oldTable);
+    kjs_fast_free(oldTable);
 
     checkConsistency();
 }
@@ -410,19 +448,21 @@ void PropertyMap::remove(const Identifier &name)
 
     // Find the thing to remove.
     unsigned h = rep->hash();
-    int i = h & _table->sizeMask;
+    int sizeMask = _table->sizeMask;
+    Entry *entries = _table->entries;
+    int i = h & sizeMask;
     int k = 0;
 #if DUMP_STATISTICS
     ++numProbes;
     ++numRemoves;
-    numCollisions += _table->entries[i].key && _table->entries[i].key != rep;
+    numCollisions += entries[i].key && entries[i].key != rep;
 #endif
-    while ((key = _table->entries[i].key)) {
+    while ((key = entries[i].key)) {
         if (rep == key)
             break;
         if (k == 0)
-            k = 1 | (h % _table->sizeMask);
-        i = (i + k) & _table->sizeMask;
+            k = 1 | (h % sizeMask);
+        i = (i + k) & sizeMask;
 #if DUMP_STATISTICS
         ++numRehashes;
 #endif
@@ -436,12 +476,16 @@ void PropertyMap::remove(const Identifier &name)
     key->deref();
     key = &UString::Rep::null;
     key->ref();
-    _table->entries[i].key = key;
-    _table->entries[i].value = 0;
-    _table->entries[i].attributes = DontEnum;
+    entries[i].key = key;
+    entries[i].value = 0;
+    entries[i].attributes = DontEnum;
     assert(_table->keyCount >= 1);
     --_table->keyCount;
+    ++_table->sentinelCount;
     
+    if (_table->sentinelCount * 4 >= _table->size)
+        rehash();
+
     checkConsistency();
 }
 
@@ -458,15 +502,12 @@ void PropertyMap::mark() const
         return;
     }
 
-    for (int i = 0; i != _table->size; ++i) {
-        UString::Rep *key = _table->entries[i].key;
-        if (key) {
-            ValueImp *v = _table->entries[i].value;
-            // Check v against 0 to handle deleted elements
-            // without comparing key to UString::Rep::null.
-            if (v && !v->marked())
-                v->mark();
-        }
+    int size = _table->size;
+    Entry *entries = _table->entries;
+    for (int i = 0; i != size; ++i) {
+        ValueImp *v = entries[i].value;
+        if (v && !v->marked())
+            v->mark();
     }
 }
 
@@ -502,8 +543,10 @@ void PropertyMap::addEnumerablesToReferenceList(ReferenceList &list, const Objec
 
     // Get pointers to the enumerable entries in the buffer.
     Entry **p = sortedEnumerables;
-    for (int i = 0; i != _table->size; ++i) {
-        Entry *e = &_table->entries[i];
+    int size = _table->size;
+    Entry *entries = _table->entries;
+    for (int i = 0; i != size; ++i) {
+        Entry *e = &entries[i];
         if (e->key && !(e->attributes & DontEnum))
             *p++ = e;
     }
@@ -537,8 +580,10 @@ void PropertyMap::addSparseArrayPropertiesToReferenceList(ReferenceList &list, c
         return;
     }
 
-    for (int i = 0; i != _table->size; ++i) {
-        UString::Rep *key = _table->entries[i].key;
+    int size = _table->size;
+    Entry *entries = _table->entries;
+    for (int i = 0; i != size; ++i) {
+        UString::Rep *key = entries[i].key;
         if (key && key != &UString::Rep::null)
         {
             UString k(key);
@@ -560,8 +605,10 @@ void PropertyMap::save(SavedProperties &p) const
             ++count;
 #endif
     } else {
-        for (int i = 0; i != _table->size; ++i)
-            if (_table->entries[i].key && !(_table->entries[i].attributes & (ReadOnly | Function)))
+        int size = _table->size;
+        Entry *entries = _table->entries;
+        for (int i = 0; i != size; ++i)
+            if (entries[i].key && !(entries[i].attributes & (ReadOnly | Function)))
                 ++count;
     }
 
@@ -601,8 +648,10 @@ void PropertyMap::save(SavedProperties &p) const
 
         // Get pointers to the entries in the buffer.
         Entry **p = sortedEntries;
-        for (int i = 0; i != _table->size; ++i) {
-            Entry *e = &_table->entries[i];
+        int size = _table->size;
+        Entry *entries = _table->entries;
+        for (int i = 0; i != size; ++i) {
+            Entry *e = &entries[i];
             if (e->key && !(e->attributes & (ReadOnly | Function)))
                 *p++ = e;
         }
@@ -641,10 +690,15 @@ void PropertyMap::checkConsistency()
         return;
 
     int count = 0;
+    int sentinelCount = 0;
     for (int j = 0; j != _table->size; ++j) {
         UString::Rep *rep = _table->entries[j].key;
-        if (!rep || rep == &UString::Rep::null)
+        if (!rep)
             continue;
+        if (rep == &UString::Rep::null) {
+            ++sentinelCount;
+            continue;
+        }
         unsigned h = rep->hash();
         int i = h & _table->sizeMask;
 	int k = 0;
@@ -656,9 +710,10 @@ void PropertyMap::checkConsistency()
 	    i = (i + k) & _table->sizeMask;
         }
         assert(i == j);
-        count++;
+        ++count;
     }
     assert(count == _table->keyCount);
+    assert(sentinelCount == _table->sentinelCount);
     assert(_table->size >= 16);
     assert(_table->sizeMask);
     assert(_table->size == _table->sizeMask + 1);
