@@ -2,7 +2,7 @@
 /*
  *  This file is part of the KDE libraries
  *  Copyright (C) 1999-2000 Harri Porten (porten@kde.org)
- *  Copyright (C) 2003 Apple Computer, Inc.
+ *  Copyright (C) 2004 Apple Computer, Inc.
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Library General Public
@@ -35,6 +35,7 @@
 #include <strings.h>
 #endif
 
+#include "fast_malloc.h"
 #include "ustring.h"
 #include "operations.h"
 #include "identifier.h"
@@ -42,10 +43,9 @@
 #include "dtoa.h"
 
 #if APPLE_CHANGES
-// malloc_good_size is not prototyped anywhere!
-extern "C" {
-  size_t malloc_good_size(size_t size);
-}
+
+#include <unicode/uchar.h>
+
 #endif
 
 namespace KJS {
@@ -138,27 +138,38 @@ bool KJS::operator==(const KJS::CString& c1, const KJS::CString& c2)
   return len == c2.size() && (len == 0 || memcmp(c1.c_str(), c2.c_str(), len) == 0);
 }
 
+// Hack here to avoid a global with a constructor; point to an unsigned short instead of a UChar.
+static unsigned short almostUChar;
+static UChar *const nonNullUCharPointer = reinterpret_cast<UChar *>(&almostUChar);
 UString::Rep UString::Rep::null = { 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0 };
-UString::Rep UString::Rep::empty = { 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0 };
+UString::Rep UString::Rep::empty = { 0, 0, 1, 0, 0, 0, nonNullUCharPointer, 0, 0, 0, 0 };
 const int normalStatBufferSize = 4096;
 static char *statBuffer = 0;
 static int statBufferSize = 0;
 
 UChar UChar::toLower() const
 {
+#if APPLE_CHANGES
+  return static_cast<unsigned short>(u_tolower(uc));
+#else
   // ### properly support unicode tolower
   if (uc >= 256 || islower(uc))
     return *this;
 
   return (unsigned char)tolower(uc);
+#endif
 }
 
 UChar UChar::toUpper() const
 {
+#if APPLE_CHANGES
+  return static_cast<unsigned short>(u_toupper(uc));
+#else
   if (uc >= 256 || isupper(uc))
     return *this;
 
   return (unsigned char)toupper(uc);
+#endif
 }
 
 UCharReference& UCharReference::operator=(UChar c)
@@ -180,6 +191,15 @@ UChar& UCharReference::ref() const
   }
 }
 
+UString::Rep *UString::Rep::createCopying(const UChar *d, int l)
+{
+  int sizeInBytes = l * sizeof(UChar);
+  UChar *copyD = static_cast<UChar *>(kjs_fast_malloc(sizeInBytes));
+  memcpy(copyD, d, sizeInBytes);
+
+  return create(copyD, l);
+}
+
 UString::Rep *UString::Rep::create(UChar *d, int l)
 {
   Rep *r = new Rep;
@@ -233,7 +253,7 @@ void UString::Rep::destroy()
   if (baseString) {
     baseString->deref();
   } else {
-    free(buf);
+    kjs_fast_free(buf);
   }
   delete this;
 }
@@ -315,9 +335,6 @@ unsigned UString::Rep::computeHash(const char *s)
 inline int UString::expandedSize(int size, int otherSize) const
 {
   int s = (size * 11 / 10) + 1 + otherSize;
-#if APPLE_CHANGES
-  s = malloc_good_size(s * sizeof(UChar)) / sizeof(UChar);
-#endif
   return s;
 }
 
@@ -337,7 +354,7 @@ void UString::expandCapacity(int requiredLength)
 
   if (requiredLength > r->capacity) {
     int newCapacity = expandedSize(requiredLength, r->preCapacity);
-    r->buf = static_cast<UChar *>(realloc(r->buf, newCapacity * sizeof(UChar)));
+    r->buf = static_cast<UChar *>(kjs_fast_realloc(r->buf, newCapacity * sizeof(UChar)));
     r->capacity = newCapacity - r->preCapacity;
   }
   if (requiredLength > r->usedCapacity) {
@@ -353,9 +370,9 @@ void UString::expandPreCapacity(int requiredPreCap)
     int newCapacity = expandedSize(requiredPreCap, r->capacity);
     int delta = newCapacity - r->capacity - r->preCapacity;
 
-    UChar *newBuf = static_cast<UChar *>(malloc(newCapacity * sizeof(UChar)));
+    UChar *newBuf = static_cast<UChar *>(kjs_fast_malloc(newCapacity * sizeof(UChar)));
     memcpy(newBuf + delta, r->buf, (r->capacity + r->preCapacity) * sizeof(UChar));
-    free(r->buf);
+    kjs_fast_free(r->buf);
     r->buf = newBuf;
 
     r->preCapacity = newCapacity - r->capacity;
@@ -373,7 +390,7 @@ UString::UString()
 
 UString::UString(char c)
 {
-    UChar *d = static_cast<UChar *>(malloc(sizeof(UChar)));
+    UChar *d = static_cast<UChar *>(kjs_fast_malloc(sizeof(UChar)));
     d[0] = c;
     rep = Rep::create(d, 1);
 }
@@ -389,7 +406,7 @@ UString::UString(const char *c)
     attach(&Rep::empty);
     return;
   }
-  UChar *d = static_cast<UChar *>(malloc(sizeof(UChar) * length));
+  UChar *d = static_cast<UChar *>(kjs_fast_malloc(sizeof(UChar) * length));
   for (int i = 0; i < length; i++)
     d[i].uc = c[i];
   rep = Rep::create(d, length);
@@ -401,9 +418,7 @@ UString::UString(const UChar *c, int length)
     attach(&Rep::empty);
     return;
   }
-  UChar *d = static_cast<UChar *>(malloc(sizeof(UChar) *length));
-  memcpy(d, c, length * sizeof(UChar));
-  rep = Rep::create(d, length);
+  rep = Rep::createCopying(c, length);
 }
 
 UString::UString(UChar *c, int length, bool copy)
@@ -412,13 +427,11 @@ UString::UString(UChar *c, int length, bool copy)
     attach(&Rep::empty);
     return;
   }
-  UChar *d;
   if (copy) {
-    d = static_cast<UChar *>(malloc(sizeof(UChar) * length));
-    memcpy(d, c, length * sizeof(UChar));
-  } else
-    d = c;
-  rep = Rep::create(d, length);
+    rep = Rep::createCopying(c, length);
+  } else {
+    rep = Rep::create(c, length);
+  }
 }
 
 UString::UString(const UString &a, const UString &b)
@@ -458,7 +471,7 @@ UString::UString(const UString &a, const UString &b)
   } else {
     // a does not qualify for append, and b does not qualify for prepend, gotta make a whole new string
     int newCapacity = expandedSize(length, 0);
-    UChar *d = static_cast<UChar *>(malloc(sizeof(UChar) * newCapacity));
+    UChar *d = static_cast<UChar *>(kjs_fast_malloc(sizeof(UChar) * newCapacity));
     memcpy(d, a.data(), aSize * sizeof(UChar));
     memcpy(d + aSize, b.data(), bSize * sizeof(UChar));
     rep = Rep::create(d, length);
@@ -593,6 +606,41 @@ UString UString::from(double d)
   return UString(buf);
 }
 
+UString UString::spliceSubstringsWithSeparators(const Range *substringRanges, int rangeCount, const UString *separators, int separatorCount) const
+{
+  int totalLength = 0;
+
+  for (int i = 0; i < rangeCount; i++) {
+    totalLength += substringRanges[i].length;
+  }
+  for (int i = 0; i < separatorCount; i++) {
+    totalLength += separators[i].size();
+  }
+
+  UChar *buffer = static_cast<UChar *>(kjs_fast_malloc(totalLength * sizeof(UChar)));
+
+  int maxCount = MAX(rangeCount, separatorCount);
+  int bufferPos = 0;
+  for (int i = 0; i < maxCount; i++) {
+    if (i < rangeCount) {
+      memcpy(buffer + bufferPos, data() + substringRanges[i].position, substringRanges[i].length * sizeof(UChar));
+      bufferPos += substringRanges[i].length;
+    }
+    if (i < separatorCount) {
+      memcpy(buffer + bufferPos, separators[i].data(), separators[i].size() * sizeof(UChar));
+      bufferPos += separators[i].size();
+    }
+  }
+
+  UString::Rep *rep = UString::Rep::create(buffer, totalLength);
+  UString result = UString(rep);
+  rep->deref();
+
+  return result;
+}
+
+
+
 UString &UString::append(const UString &t)
 {
   int thisSize = size();
@@ -614,7 +662,7 @@ UString &UString::append(const UString &t)
     rep->_hash = 0;
   } else if (thisOffset + thisSize == usedCapacity()) {
     // this reaches the end of the buffer - extend it
-    expandCapacity(length);
+    expandCapacity(thisOffset + length);
     memcpy(const_cast<UChar *>(data() + thisSize), t.data(), tSize * sizeof(UChar));
     Rep *newRep = Rep::create(rep, 0, length);
     release();
@@ -622,7 +670,7 @@ UString &UString::append(const UString &t)
   } else {
     // this is shared with someone using more capacity, gotta make a whole new string
     int newCapacity = expandedSize(length, 0);
-    UChar *d = static_cast<UChar *>(malloc(sizeof(UChar) * newCapacity));
+    UChar *d = static_cast<UChar *>(kjs_fast_malloc(sizeof(UChar) * newCapacity));
     memcpy(d, data(), thisSize * sizeof(UChar));
     memcpy(const_cast<UChar *>(d + thisSize), t.data(), tSize * sizeof(UChar));
     release();
@@ -666,7 +714,7 @@ UString &UString::append(const char *t)
   } else {
     // this is shared with someone using more capacity, gotta make a whole new string
     int newCapacity = expandedSize(length, 0);
-    UChar *d = static_cast<UChar *>(malloc(sizeof(UChar) * newCapacity));
+    UChar *d = static_cast<UChar *>(kjs_fast_malloc(sizeof(UChar) * newCapacity));
     memcpy(d, data(), thisSize * sizeof(UChar));
     for (int i = 0; i < tSize; ++i)
       d[thisSize+i] = t[i];
@@ -687,7 +735,7 @@ UString &UString::append(unsigned short c)
   if (length == 0) {
     // this is empty - must make a new rep because we don't want to pollute the shared empty one 
     int newCapacity = expandedSize(1, 0);
-    UChar *d = static_cast<UChar *>(malloc(sizeof(UChar) * newCapacity));
+    UChar *d = static_cast<UChar *>(kjs_fast_malloc(sizeof(UChar) * newCapacity));
     d[0] = c;
     release();
     rep = Rep::create(d, 1);
@@ -710,7 +758,7 @@ UString &UString::append(unsigned short c)
   } else {
     // this is shared with someone using more capacity, gotta make a whole new string
     int newCapacity = expandedSize((length + 1), 0);
-    UChar *d = static_cast<UChar *>(malloc(sizeof(UChar) * newCapacity));
+    UChar *d = static_cast<UChar *>(kjs_fast_malloc(sizeof(UChar) * newCapacity));
     memcpy(d, data(), length * sizeof(UChar));
     d[length] = c;
     release();
@@ -772,7 +820,7 @@ UString &UString::operator=(const char *c)
     rep->_hash = 0;
   } else {
     release();
-    d = static_cast<UChar *>(malloc(sizeof(UChar) * l));
+    d = static_cast<UChar *>(kjs_fast_malloc(sizeof(UChar) * l));
     rep = Rep::create(d, l);
   }
   for (int i = 0; i < l; i++)
@@ -999,8 +1047,10 @@ int UString::find(const UString &f, int pos) const
   const UChar *end = data() + sz - fsz;
   long fsizeminusone = (fsz - 1) * sizeof(UChar);
   const UChar *fdata = f.data();
+  unsigned short fchar = fdata->uc;
+  ++fdata;
   for (const UChar *c = data() + pos; c <= end; c++)
-    if (*c == *fdata && !memcmp(c + 1, fdata + 1, fsizeminusone))
+    if (c->uc == fchar && !memcmp(c + 1, fdata, fsizeminusone))
       return (c-data());
 
   return -1;
@@ -1056,14 +1106,19 @@ int UString::rfind(UChar ch, int pos) const
 
 UString UString::substr(int pos, int len) const
 {
+  int s = size();
+
   if (pos < 0)
     pos = 0;
-  else if (pos >= (int) size())
-    pos = size();
+  else if (pos >= s)
+    pos = s;
   if (len < 0)
-    len = size();
-  if (pos + len >= (int) size())
-    len = size() - pos;
+    len = s;
+  if (pos + len >= s)
+    len = s - pos;
+
+  if (pos == 0 && len == s)
+    return *this;
 
   UString::Rep *newRep = Rep::create(rep, pos, len);
   UString result(newRep);
@@ -1072,28 +1127,17 @@ UString UString::substr(int pos, int len) const
   return result;
 }
 
-void UString::attach(Rep *r)
-{
-  rep = r;
-  rep->ref();
-}
-
 void UString::detach()
 {
   if (rep->rc > 1 || rep->baseString) {
     int l = size();
-    UChar *n = static_cast<UChar *>(malloc(sizeof(UChar) * l));
+    UChar *n = static_cast<UChar *>(kjs_fast_malloc(sizeof(UChar) * l));
     memcpy(n, data(), l * sizeof(UChar));
     release();
     rep = Rep::create(n, l);
   }
 }
 
-void UString::release()
-{
-  rep->deref();
-}
-
 bool KJS::operator==(const UString& s1, const UString& s2)
 {
   if (s1.rep->len != s2.rep->len)
@@ -1283,7 +1327,7 @@ CString UString::UTF8String() const
     } else if (c < 0x800) {
       *p++ = (char)((c >> 6) | 0xC0); // C0 is the 2-byte flag for UTF-8
       *p++ = (char)((c | 0x80) & 0xBF); // next 6 bits, with high bit set
-    } else if (c >= 0xD800 && c <= 0xDBFF && i < length && d[i+1].uc >= 0xDC00 && d[i+2].uc <= 0xDFFF) {
+    } else if (c >= 0xD800 && c <= 0xDBFF && i < length && d[i+1].uc >= 0xDC00 && d[i+1].uc <= 0xDFFF) {
       unsigned sc = 0x10000 + (((c & 0x3FF) << 10) | (d[i+1].uc & 0x3FF));
       *p++ = (char)((sc >> 18) | 0xF0); // F0 is the 4-byte flag for UTF-8
       *p++ = (char)(((sc >> 12) | 0x80) & 0xBF); // next 6 bits, with high bit set
@@ -1305,166 +1349,4 @@ CString UString::UTF8String() const
   return result;
 }
 
-struct StringOffset {
-    int offset;
-    int locationInOffsetsArray;
-};
-
-static int compareStringOffsets(const void *a, const void *b)
-{
-    const StringOffset *oa = static_cast<const StringOffset *>(a);
-    const StringOffset *ob = static_cast<const StringOffset *>(b);
-    
-    if (oa->offset < ob->offset) {
-        return -1;
-    }
-    if (oa->offset > ob->offset) {
-        return +1;
-    }
-    return 0;
-}
-
-const int sortedOffsetsFixedBufferSize = 128;
-
-static StringOffset *createSortedOffsetsArray(const int offsets[], int numOffsets,
-    StringOffset sortedOffsetsFixedBuffer[sortedOffsetsFixedBufferSize])
-{
-    // Allocate the sorted offsets.
-    StringOffset *sortedOffsets;
-    if (numOffsets <= sortedOffsetsFixedBufferSize) {
-        sortedOffsets = sortedOffsetsFixedBuffer;
-    } else {
-        sortedOffsets = new StringOffset [numOffsets];
-    }
-
-    // Copy offsets and sort them.
-    // (Since qsort showed up on profiles, hand code for numbers up to 3.)
-
-    switch (numOffsets) {
-        case 0:
-            break;
-        case 1:
-            sortedOffsets[0].offset = offsets[0];
-            sortedOffsets[0].locationInOffsetsArray = 0;
-            break;
-        case 2: {
-            if (offsets[0] <= offsets[1]) {
-                sortedOffsets[0].offset = offsets[0];
-                sortedOffsets[0].locationInOffsetsArray = 0;
-                sortedOffsets[1].offset = offsets[1];
-                sortedOffsets[1].locationInOffsetsArray = 1;
-            } else {
-                sortedOffsets[0].offset = offsets[1];
-                sortedOffsets[0].locationInOffsetsArray = 1;
-                sortedOffsets[1].offset = offsets[0];
-                sortedOffsets[1].locationInOffsetsArray = 0;
-            }
-            break;
-        }
-        case 3: {
-            int i0, i1, i2;
-            if (offsets[0] <= offsets[1]) {
-                if (offsets[0] <= offsets[2]) {
-                    i0 = 0;
-                    if (offsets[1] <= offsets[2]) {
-                        i1 = 1; i2 = 2;
-                    } else {
-                        i1 = 2; i2 = 1;
-                    }
-                } else {
-                    i0 = 2; i1 = 0; i2 = 1;
-                }
-            } else {
-                if (offsets[1] <= offsets[2]) {
-                    i0 = 1;
-                    if (offsets[0] <= offsets[2]) {
-                        i1 = 0; i2 = 2;
-                    } else {
-                        i1 = 2; i2 = 0;
-                    }
-                } else {
-                    i0 = 2; i1 = 1; i2 = 0;
-                }
-            }
-            sortedOffsets[0].offset = offsets[i0];
-            sortedOffsets[0].locationInOffsetsArray = i0;
-            sortedOffsets[1].offset = offsets[i1];
-            sortedOffsets[1].locationInOffsetsArray = i1;
-            sortedOffsets[2].offset = offsets[i2];
-            sortedOffsets[2].locationInOffsetsArray = i2;
-            break;
-        }
-        default:
-            for (int i = 0; i != numOffsets; ++i) {
-                sortedOffsets[i].offset = offsets[i];
-                sortedOffsets[i].locationInOffsetsArray = i;
-            }
-            qsort(sortedOffsets, numOffsets, sizeof(StringOffset), compareStringOffsets);
-    }
-
-    return sortedOffsets;
-}
-
-// Note: This function assumes valid UTF-8.
-// It can even go into an infinite loop if the passed in string is not valid UTF-8.
-void convertUTF16OffsetsToUTF8Offsets(const char *s, int *offsets, int numOffsets)
-{
-    // Allocate buffer.
-    StringOffset fixedBuffer[sortedOffsetsFixedBufferSize];
-    StringOffset *sortedOffsets = createSortedOffsetsArray(offsets, numOffsets, fixedBuffer);
-
-    // Walk through sorted offsets and string, adjusting all the offests.
-    // Offsets that are off the ends of the string map to the edges of the string.
-    int UTF16Offset = 0;
-    const char *p = s;
-    for (int oi = 0; oi != numOffsets; ++oi) {
-        const int nextOffset = sortedOffsets[oi].offset;
-        while (*p && UTF16Offset < nextOffset) {
-            // Skip to the next character.
-            const int sequenceLength = inlineUTF8SequenceLength(*p);
-            assert(sequenceLength >= 1 && sequenceLength <= 4);
-            p += sequenceLength;
-            // Characters that take a 4 byte sequence in UTF-8 take two bytes in UTF-16.
-            UTF16Offset += sequenceLength < 4 ? 1 : 2;
-        }
-        offsets[sortedOffsets[oi].locationInOffsetsArray] = p - s;
-    }
-
-    // Free buffer.
-    if (sortedOffsets != fixedBuffer) {
-        delete [] sortedOffsets;
-    }
-}
-
-// Note: This function assumes valid UTF-8.
-// It can even go into an infinite loop if the passed in string is not valid UTF-8.
-void convertUTF8OffsetsToUTF16Offsets(const char *s, int *offsets, int numOffsets)
-{
-    // Allocate buffer.
-    StringOffset fixedBuffer[sortedOffsetsFixedBufferSize];
-    StringOffset *sortedOffsets = createSortedOffsetsArray(offsets, numOffsets, fixedBuffer);
-
-    // Walk through sorted offsets and string, adjusting all the offests.
-    // Offsets that are off the end of the string map to the edges of the string.
-    int UTF16Offset = 0;
-    const char *p = s;
-    for (int oi = 0; oi != numOffsets; ++oi) {
-        const int nextOffset = sortedOffsets[oi].offset;
-        while (*p && (p - s) < nextOffset) {
-            // Skip to the next character.
-            const int sequenceLength = inlineUTF8SequenceLength(*p);
-            assert(sequenceLength >= 1 && sequenceLength <= 4);
-            p += sequenceLength;
-            // Characters that take a 4 byte sequence in UTF-8 take two bytes in UTF-16.
-            UTF16Offset += sequenceLength < 4 ? 1 : 2;
-        }
-        offsets[sortedOffsets[oi].locationInOffsetsArray] = UTF16Offset;
-    }
-
-    // Free buffer.
-    if (sortedOffsets != fixedBuffer) {
-        delete [] sortedOffsets;
-    }
-}
-
 } // namespace KJS
