@@ -2,7 +2,7 @@
 /*
  *  This file is part of the KDE libraries
  *  Copyright (C) 1999-2000 Harri Porten (porten@kde.org)
- *  Copyright (C) 2003 Apple Computer, Inc.
+ *  Copyright (C) 2004 Apple Computer, Inc.
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Library General Public
@@ -24,6 +24,8 @@
 #ifndef _KJS_USTRING_H_
 #define _KJS_USTRING_H_
 
+#include "fast_malloc.h"
+
 #if APPLE_CHANGES
 #include <sys/types.h>
 #ifndef KWQ_UNSIGNED_TYPES_DEFINED
@@ -206,6 +208,7 @@ namespace KJS {
       friend bool operator==(const UString&, const UString&);
       
       static Rep *create(UChar *d, int l);
+      static Rep *createCopying(const UChar *d, int l);
       static Rep *create(Rep *base, int offset, int length);
       void destroy();
       
@@ -216,6 +219,8 @@ namespace KJS {
       static unsigned computeHash(const UChar *, int length);
       static unsigned computeHash(const char *);
 
+      KJS_FAST_ALLOCATED;
+
       void ref() { ++rc; }
       void deref() { if (--rc == 0) destroy(); }
 
@@ -289,6 +294,8 @@ namespace KJS {
      */
     ~UString() { release(); }
 
+    KJS_FAST_ALLOCATED;
+
     /**
      * Constructs a string from an int.
      */
@@ -306,6 +313,16 @@ namespace KJS {
      */
     static UString from(double d);
 
+    struct Range {
+    public:
+      Range(int pos, int len) : position(pos), length(len) {}
+      Range() {}
+      int position;
+      int length;
+    };
+
+    UString spliceSubstringsWithSeparators(const Range *substringRanges, int rangeCount, const UString *separators, int separatorCount) const;
+
     /**
      * Append another string.
      */
@@ -453,9 +470,9 @@ namespace KJS {
 #endif
   private:
     UString(Rep *r) { attach(r); }
-    void attach(Rep *r);
+    void attach(Rep *r) { rep = r; r->ref(); }
     void detach();
-    void release();
+    void release() { rep->deref(); }
     int expandedSize(int size, int otherSize) const;
     int usedCapacity() const;
     int usedPreCapacity() const;
@@ -500,16 +517,6 @@ namespace KJS {
   // Returns -1 if the sequence is not valid (including presence of extra bytes).
   int decodeUTF8Sequence(const char *);
 
-  // Given a UTF-8 string, converts offsets from the UTF-16 form of the string into offsets into the UTF-8 string.
-  // Note: This function can overrun the buffer if the string contains a partial UTF-8 sequence, so it should
-  // not be called with strings that might contain such sequences.
-  void convertUTF16OffsetsToUTF8Offsets(const char *UTF8String, int *offsets, int numOffsets);
-
-  // Given a UTF-8 string, converts offsets from the UTF-8 string into offsets into the UTF-16 form of the string.
-  // Note: This function can overrun the buffer if the string contains a partial UTF-8 sequence, so it should
-  // not be called with strings that might contain such sequences.
-  void convertUTF8OffsetsToUTF16Offsets(const char *UTF8String, int *offsets, int numOffsets);
-
 }; // namespace
 
 #endif
