@@ -20,21 +20,22 @@
  * 
  */
 /**
-    File Name:          15.9.5.7.js
-    ECMA Section:       15.9.5.7
-    Description:        Date.prototype.getUTCFullYear
+    File Name:          15.9.5.19.js
+    ECMA Section:       15.9.5.19
+    Description:        Date.prototype.getUTCSeconds
+
+    1.  Let t be this time value.
+    2.  If t is NaN, return NaN.
+    3.  Return SecFromTime(t).
 
-    1.Let t be this time value.
-    2.If t is NaN, return NaN.
-    3.Return YearFromTime(t).
     Author:             christine@netscape.com
     Date:               12 november 1997
 */
 
-    var SECTION = "15.9.5.7";
+    var SECTION = "15.9.5.19";
     var VERSION = "ECMA_1";
     startTest();
-    var TITLE   = "Date.prototype.getUTCFullYear()";
+    var TITLE   = "Date.prototype.getUTCSeconds()";
 
     writeHeaderToLog( SECTION + " "+ TITLE);
 
@@ -67,41 +68,23 @@
     addTestCase( UTC_JAN_1_2005 );
 
     testcases[tc++] = new TestCase( SECTION,
-                                    "(new Date(NaN)).getUTCFullYear()",
+                                    "(new Date(NaN)).getUTCSeconds()",
                                     NaN,
-                                    (new Date(NaN)).getUTCFullYear() );
+                                    (new Date(NaN)).getUTCSeconds() );
 
     testcases[tc++] = new TestCase( SECTION,
-                                    "Date.prototype.getUTCFullYear.length",
+                                    "Date.prototype.getUTCSeconds.length",
                                     0,
-                                    Date.prototype.getUTCFullYear.length );
-
+                                    Date.prototype.getUTCSeconds.length );
     test();
 function addTestCase( t ) {
-    testcases[tc++] = new TestCase( SECTION,
-                                    "(new Date("+t+")).getUTCFullYear()",
-                                    YearFromTime(t),
-                                    (new Date(t)).getUTCFullYear() );
-
-    testcases[tc++] = new TestCase( SECTION,
-                                    "(new Date("+(t+1)+")).getUTCFullYear()",
-                                    YearFromTime(t+1),
-                                    (new Date(t+1)).getUTCFullYear() );
-
-    testcases[tc++] = new TestCase( SECTION,
-                                    "(new Date("+(t-1)+")).getUTCFullYear()",
-                                    YearFromTime(t-1),
-                                    (new Date(t-1)).getUTCFullYear() );
-
-    testcases[tc++] = new TestCase( SECTION,
-                                    "(new Date("+(t-TZ_ADJUST)+")).getUTCFullYear()",
-                                    YearFromTime(t-TZ_ADJUST),
-                                    (new Date(t-TZ_ADJUST)).getUTCFullYear() );
-
-    testcases[tc++] = new TestCase( SECTION,
-                                    "(new Date("+(t+TZ_ADJUST)+")).getUTCFullYear()",
-                                    YearFromTime(t+TZ_ADJUST),
-                                    (new Date(t+TZ_ADJUST)).getUTCFullYear() );
+    for ( m = 0; m <= 60; m+=10 ) {
+            t += 1000;
+            testcases[tc++] = new TestCase( SECTION,
+                                    "(new Date("+t+")).getUTCSeconds()",
+                                    SecFromTime(t),
+                                    (new Date(t)).getUTCSeconds() );
+    }
 }
 function test() {
     for ( tc=0; tc < testcases.length; tc++ ) {
